from repoze.formapi.parser import parse
from repoze.formapi.parser import missing

import types
import re

def get_instances_of(type, *bases):
    for base in bases:
        for value in base.__dict__.values():
            if isinstance(value, type):
                yield value
        for value in get_instances_of(type, *base.__bases__):
            yield value

class Validator(object):
    """Wrapper for validators.

    This calls a validator object (usually a method) and collects all it's
    errors. It also sets a flag that the form library can use to know that it
    is a validator."""

    def __init__(self, func, *fields):
        self.func = func
        # Fields can be empty, in that case we want to have and empty path
        if not fields:
            self.fieldpaths = ((),)
        else:
            self.fieldpaths = [f.split('.') for f in fields]

    def __call__(self, form):
        for error in self.func(form):
            for fieldpath in self.fieldpaths:
                yield (fieldpath, error)

class Action(object):
    def __init__(self, action, name=None, submitted=False):
        self.action = action
        self.name = name
        self.submitted = submitted

    @property
    def __call__(self):
        return self.action

    def __nonzero__(self):
        return self.submitted

    def __repr__(self):
        return '<%s name="%s" submitted="%s">' % (
            type(self).__name__, self.name or "", str(bool(self.submitted)))

class metaclass(type):
    def __init__(kls, name, bases, dict):
        kls.validators = tuple(get_instances_of(Validator, kls))
        kls.actions = tuple(get_instances_of(Action, kls))

class Form(object):
    """Base form class. Optionally pass a dictionary as ``data`` and a
    WebOb-like request object as ``request``."""

    __metaclass__ = metaclass

    fields = {}
    status = None
    prefix = None
    action = None

    def __init__(self, data=None, context=None, request=None, params=None, prefix=None):
        self.context = context
        self.request = request

        if context is not None:
            if data is not None:
                raise ValueError(
                    "Can't provide both ``data`` and ``context``.")

            # proxy the context object
            data = Proxy(context)

        self.data = Data(data)

        if prefix is None:
            prefix = self.prefix

        if request is not None:
            if params is not None:
                raise ValueError(
                    "Can't provide both ``params`` and ``request``.")
            params = request.params.items()

        # find action parameters
        action_params = {}
        if prefix is not None and params is not None:
            re_prefix = re.compile(r'^%s[._-](?P<name>.*)' % prefix)
            for key, value in params:
                if key == prefix:
                    action_params[None] = value
                else:
                    m = re_prefix.search(key)
                    if m is not None:
                        action_params[m.group('name')] = value

        # initialize form actions
        actions = self.actions = []
        for action in type(self).actions:
            name = action.name
            action = Action(action.__call__, name, name in action_params)
            actions.append(action)
            if action:
                self.action = action

        # conditionally apply request parameters if:
        # 1. no prefix has been set
        # 2. there is a submitted action
        # 3. there are no defined actions, but a default action was submitted
        if params is not None and (
            prefix is None or \
            filter(None, actions) or \
            len(actions) == 0 and action_params.get(None) is not None):
            params = list(params)
        else:
            params = ()

        # Parse parameter input
        data, errors = parse(params, self.fields)

        if len(params):
            self.data.update(data)

        self.errors = errors
        self.prefix = prefix

    def __call__(self):
        """Calls the first submitted action and returns the value."""

        if self.action is not None:
            self.status = self.action(self, self.data)
        return self.status

    def validate(self):
        """Validates the request against the form fields. Returns
        ``True`` if all fields validate, else ``False``."""
        for validator in self.validators:
            for field_path, validation_error in validator(self):
                errors = self.errors
                for field in field_path:
                    errors = errors[field]
                errors += validation_error

        return not bool(self.errors)

class ValidationError(Exception):
    """Represents a field validation error."""

    def __init__(self, field, msg):
        if not isinstance(msg, unicode):
            msg = unicode(msg)
        self.field = field
        self.msg = msg

    def __repr__(self):
        return '<%s field="%s" %s>' % (
            type(self).__name__, self.field, repr(self.msg))

    def __str__(self):
        return str(unicode(self))

    def __unicode__(self):
        return self.msg

class Data(list):
    """Form data object with dictionary-like interface. If initialized
    with a ``data`` object, this will be used to provide default
    values, if not set in the ``request``. Updates to the object are
    transient until the ``save`` method is invoked."""

    def __init__(self, data):
        if data is not None:
            self.append(data)
        self.append({})

    def __getitem__(self, name):
        for data in reversed(self):
            try:
                value = data[name]
            except KeyError:
                continue

            if value is not missing:
                return value

    def __setitem__(self, name, value):
        self.tail[name] = value

    @property
    def tail(self):
        return list.__getitem__(self, -1)

    @property
    def head(self):
        return list.__getitem__(self, 0)

    def update(self, data):
        """Updates the dictionary by appending ``data`` to the list at
        the position just before the current dictionary."""

        self.insert(-1, data)

    def save(self):
        """Flattens the dictionary, saving changes to the data object."""

        while len(self) > 1:
            for name, value in self.pop(1).items():
                self.head[name] = value
        self.append({})

class Proxy(object):
    """Proxy object; reads and writes to attributes are forwarded to
    the provided object. Descriptors are supported: they must read and
    write to ``self.context``. Note, that all attribute names are
    supported, including the name 'context'.

    Any object can be the context of a proxy.

    >>> class Content(object):
    ...     pass
    
    >>> from repoze.formapi import Proxy
    
    >>> class ContentProxy(Proxy):
    ...     def get_test_descriptor(self):
    ...         return self.test_descriptor
    ...
    ...     def set_test_descriptor(self, value):
    ...         self.test_descriptor = value + 1
    ...
    ...     def get_get_only(self):
    ...         return self.test_get_only
    ...
    ...     test_descriptor = property(get_test_descriptor, set_test_descriptor)
    ...     test_get_only = property(get_get_only)
    
    >>> context = Content()
    >>> proxy = ContentProxy(context)

    We can read and write to the ``context`` attribute.
    
    >>> proxy.test = 42
    >>> proxy.test
    42

    Descriptors have access to the original context.
    
    >>> proxy.test_descriptor = 41
    >>> proxy.test_descriptor
    42

    Descriptors that only define a getter are supported.
    
    >>> proxy.test_get_only = 41
    >>> proxy.test_get_only
    41

    Proxies provide dictionary-access to attributes.

    >>> proxy['test']
    42

    >>> proxy['test'] = 41
    >>> proxy.test
    41
    """

    def __init__(self, context):
        # instantiate a base proxy object with this context
        serf = object.__new__(Proxy)
        object.__setattr__(serf, '_context', context)
        object.__setattr__(self, '_context', context)
        object.__setattr__(self, '_serf', serf)

    def __getattribute__(self, name):
        prop = object.__getattribute__(type(self), '__dict__').get(name)
        if prop is not None:
            # call getter in the context of the proxy object
            serf = object.__getattribute__(self, '_serf')
            return prop.fget(serf)
        else:
            return getattr(
                object.__getattribute__(self, '_context'), name)

    def __setattr__(self, name, value):
        prop = object.__getattribute__(type(self), '__dict__').get(name)
        if prop is not None:
            # property might be read-only (e.g. does not define a
            # setter); in this case we just set attribute on context.
            setter = prop.fset
            if setter is not None:
                # call setter in the context of the proxy object
                serf = object.__getattribute__(self, '_serf')
                return prop.fset(serf, value)
        setattr(
            object.__getattribute__(self, '_context'), name, value)

    __getitem__ = __getattribute__
    __setitem__ = __setattr__
    
def action(name):
    if isinstance(name, types.FunctionType):
        return Action(name)
    else:
        def decorator(action):
            return Action(action, name)
        return decorator

def validator(*args):
    # If the first (and only) argument is a callable process it
    if len(args) == 1 and callable(args[0]):
        return Validator(args[0])
    # Treat the args as field names and prepare a wrapper
    else:
        return lambda func: Validator(func, *args)



import pprint

from repoze.formapi.py24 import defaultdict
from repoze.formapi.error import Errors


def parse(params, fields):
    """Return ``(data, errors)`` tuple.

    This function parses, converts and validates the parameter
    sequence (name, value) pairs.

    The converted data is returned as the ``data`` value, while errors
    encountered during conversion are set in the ``errors`` dict.

    The ``fields`` parameter is used to describe the data
    structure. It's a nested dictionary that ends in simple types, or
    a list of such, e.g.:

        >>> fields = {
        ...     "user": {
        ...         "name": str,
        ...         "nick": str,
        ...         "age": int,
        ...         "extra": str,
        ...     }
        ... }

    Given a sequence of parameteres, the function returns a data
    structure that matches the fields definition.

        >>> params = (
        ...     ("user.name", "Fred Kaputnik"),
        ...     ("user.nick", "fred"),
        ...     ("user.age", 42),
        ...     ("user.extra", ""))

    Note that the field names appear as 'dotted paths'.

        >>> from repoze.formapi.parser import parse
        >>> data, errors = parse(params, fields)

    Since all the tuples in ``params`` are valid, we expect no
    validation errors::

        >>> len(errors)
        0

    We can turn the returned data into a dict:

        >>> print repr(data)
        {'user': {'age': 42, 'extra': '', 'name': 'Fred Kaputnik', 'nick': 'fred'}}

    Similarly, we can pass the data as keyword arguments:

        >>> (lambda **kwargs: True)(**data)
        True

    To get to the parsed field data, we use a similar approach.

        >>> data['user']['name']
        'Fred Kaputnik'
        >>> data['user']['nick']
        'fred'
        >>> data['user']['age']
        42
        >>> data['user']['extra']
        ''

    To illustrate error handling, we can violate the validation
    constraint of the ``age`` field.

        >>> data, errors = parse((("user.age", "ten"),), fields)

        >>> data['user']['age']
        'ten'

        >>> 'ten' in errors['user']['age'][0]
        True

    Note that the ``data`` and ``errors`` dictionaries provide a
    default value of ``missing`` for missing entries.

        >>> from repoze.formapi.parser import missing

    It evaluates to false.

        >>> bool(missing)
        False

        >>> data, errors = parse((), fields)

        >>> 'age' in data['user']
        False

        >>> data['user']['age'] is missing
        True

    The fields structure may end in a list of a simple type.

        >>> fields = {
        ...     "user": {
        ...         "friends": [str],
        ...     }
        ... }

    In this case, we expect the ``friends`` entry to be a list of strings::

        >>> params = (
        ...     ("user.friends", "stefan"),
        ...     ("user.friends", "malthe"))

        >>> data, errors = parse(params, fields)

    As expected, we get a list. Note that this list items appear in
    the order they appear in the reqeuest parameters.

        >>> data['user']['friends']
        ['stefan', 'malthe']

    Tuples are also supported.

        >>> fields = {
        ...     "points": (int,)
        ...     }

        >>> params = (("points", 42), ("points", 10))

        >>> data, errors = parse(params, fields)

        >>> data['points']
        (42, 10)

    Sequence types may only appear as end-points. The following field
    definition is invalid.

        >>> fields = {
        ...     "users": [{
        ...         "name": str,
        ...         "nick": str}]
        ...     }

        >>> params = (("users.name", "Foo"),)
        >>> parse(params, fields)
        Traceback (most recent call last):
         ...
        TypeError: Sequences are only allowed as end-points.

    Dictionary entries may be dynamic.

        >>> fields = {
        ...     "user": {
        ...         str: {
        ...            "name": str,
        ...         }
        ...     }
        ... }

        >>> params = (
        ...     ("user.stefan.name", "Stefan Eletzhofer"),
        ...     ("user.malthe.name", "Malthe Borch"))

        >>> data, errors = parse(params, fields)

    As expected, we get a list. Note that this list items appear in
    the order they appear in the reqeuest parameters.

        >>> data['user']['stefan']['name']
        'Stefan Eletzhofer'

    If a trivial input results in a conversion error, the value is set
    to ``None``.

        >>> fields = {
        ...     'user': str,
        ...     'age': int
        ...     }

        >>> params = (
        ...     ('user', ''),
        ...     ('age', ''))

        >>> data, errors = parse(params, fields)

     Empty strings are valid for string fields.

        >>> data['user']
        ''

     Empty strings are trivial for integer fields; it's regarded as a
     non-input.

        >>> data['age']

     This is not an error:

        >>> str(errors['age'])
        ''

     If the field is required, an empty string is an invalid input for
     a number. Invalid inputs are preserved as-is for required fields.

        >>> from repoze.formapi.parser import required
        >>> fields['age'] = required(int)
        >>> data, errors = parse(params, fields)

        >>> data['age'] is None
        True

        >>> bool(errors['age'])
        True

     The same goes for non-trivial invalid inputs.

        >>> data, errors = parse((('age', 'ten'),), fields)
        >>> data['age']
        'ten'

        >>> bool(errors['age'])
        True

    """

    data = Parser(fields)
    errors = Errors()

    parsed_errors = Parser(fields, coerce=False)

    for name, value in params:
        path = tuple(name.split('.'))
        try:
            data[path] = value
        except KeyError:
            continue
        except ValueError, error:
            e = errors
            for p in path:
                e = e[p]
            e += str(error)

    for path, message in parsed_errors:
        e = errors
        for p in path:
            e = e[path]
        e += message

    return data, errors


class Parser(object):
    """Form input parser.

    Any field name that comforms to the field definition may be
    resolved; if not set, an empty value is returned.

        >>> from repoze.formapi.parser import Parser
        >>> from repoze.formapi.parser import missing

        >>> parser = Parser({
        ...     'name': str,
        ...     'users': {
        ...         str: {
        ...            'username': str,
        ...            'id': int,
        ...            'groups': [str]}
        ...         }
        ...     })

    We'll first demonstrate that valid keys return an empty value.

        >>> parser['name'] is missing
        True

        >>> parser['users', 'foo', 'username'] is missing
        True

        >>> parser['users', 'foo', 'groups']
        []

    An exception is raised if an invalid key is tried.

        >>> parser['users', 'foo', 'bar']
        Traceback (most recent call last):
         ...
        KeyError: 'bar'

    The truth-value of an empty ``Parser`` instance is ``False``.

        >>> bool(parser)
        False

    We may set values for valid keys.

        >>> parser['name'] = 'Foo'
        >>> parser['users', 'foo', 'id'] = 1
        >>> parser['users', 'foo', 'username'] = 'foo'

    Sequences append on assignments.

        >>> parser['users', 'foo', 'groups'] = 'foo'
        >>> parser['users', 'foo', 'groups'] = 'bar'

    As expected, we can retrieve the values in exchange for the keys.

        >>> parser['name']
        'Foo'

        >>> parser['users', 'foo', 'username']
        'foo'

        >>> parser['users', 'foo', 'id']
        1

        >>> parser['users', 'foo', 'groups']
        ['foo', 'bar']

    Sequences are replaced, if assigned a sequence.

        >>> parser['users', 'foo', 'groups'] = ['bar']
        >>> parser['users', 'foo', 'groups']
        ['bar']

    Again, an exception is raised if an invalid key is tried.

        >>> parser['users', 'foo', 'bar'] = 'baz'
        Traceback (most recent call last):
         ...
        KeyError: 'bar'

    Values are attempted coerced to the field type.

        >>> parser['users', 'foo', 'id'] = '1'
        >>> parser['users', 'foo', 'id']
        1

    Coercing must not fail.

        >>> parser['users', 'foo', 'id'] = 'one'
        Traceback (most recent call last):
         ...
        ValueError: ...

    We may always assign ``None``.

        >>> parser['users', 'foo', 'id'] = None

    The truth-value of a non-empty ``Parser`` instance is ``True``.

        >>> bool(parser)
        True

    If the path is not exhausted, a new parser instance is
    returned, which curries the path onto new requests.

        >>> parser['users']['foo']['username']
        'foo'

    Each segment is iterable.

        >>> tuple(parser['users'])
        ('foo',)

        >>> tuple(sorted(parser['users']['foo']))
        ('groups', 'id', 'username')

    """

    def __init__(self, fields, data=None, path=(), coerce=True):
        self.fields = fields

        if data is None:
            data = {}

        self.data = data
        self.path = path
        self.coerce = coerce

    def __getitem__(self, path):
        if not isinstance(path, tuple):
            path = (path,)
        path = self.path + tuple(path)
        value = self.data.get(path, missing)
        if value is not missing:
            return value

        # verify path; we want to raise an exception if the path does
        # not comply with the field definition
        data_type = self.traverse(path)

        if isinstance(data_type, list):
            return []

        if isinstance(data_type, tuple):
            return ()

        if isinstance(data_type, dict):
            return Parser(self.fields, self.data, path, self.coerce)

        return missing

    def __setitem__(self, path, value):
        if not isinstance(path, tuple):
            path = (path,)
        key = self.path + tuple(path)

        # verify path; we want to raise an exception if the path does
        # not comply with the field definition
        data_type = self.traverse(path)

        error = False
        if isinstance(data_type, (tuple, list)):
            if isinstance(value, (tuple, list)):
                if key in self.data:
                    del self.data[key]
            else:
                value = (value,)

            for v in value:
                if v is not None and self.coerce:
                    try:
                        v = data_type[0](v)
                    except:
                        error = True

                if isinstance(data_type, list):
                    self.data.setdefault(
                        key, []).append(v)
                else:
                    if key in self.data:
                        self.data[key] += (v,)
                    else:
                        self.data[key] = (v,)
        else:
            if value is not None and self.coerce:
                try:
                    value = data_type(value)
                except KeyError:
                    raise
                except MissingError:
                    error = True
                    value = None
                except:
                    if value:
                        error = True
                    else:
                        value = None

            self.data[key] = value

        if value is not None:
            self.data[None] = True

        if error is True:
            raise

    def __nonzero__(self):
        return self.data.get(None, False)

    def __repr__(self):
        data = self.parse()
        return repr(data)

    def __iter__(self):
        items = set()
        for path in self.data:
            if path is None:
                continue

            length = len(self.path)

            if tuple(path[:length]) == self.path:
                item = path[length]
                if item not in items:
                    yield item

                items.add(item)

    def keys(self):
        return list(self)

    def items(self):
        return list(self.iteritems())

    def iteritems(self):
        for key in self:
            yield (key, self[key])

    def parse(self):
        _data = {}
        for path, value in self.data.items():
            if path is None:
                continue

            data = _data
            for key in path[:-1]:
                if key in data:
                    data = data[key]
                else:
                    data[key] = defaultdict(lambda: missing)
                    data = data[key]
            data[path[-1]] = value
        return _data

    def traverse(self, path):
        path = list(path)
        fields = self.fields
        while path:
            if not isinstance(fields, dict):
                raise TypeError(
                    "Sequences are only allowed as end-points.")

            value = path.pop(0)
            if len(fields) == 1:
                key = fields.keys()[0]
                if key in (str, unicode, int):
                    # make sure value conforms to data type
                    if not isinstance(value, key):
                        raise TypeError(
                            "Must be type '%s' (got '%s')." % (
                            (key.__name__, type(value).__name__)))
                    fields = fields[key]
                    continue
            fields = fields[value]
        return fields


def required(cls, msg="Required field"):
    class required(cls):
        def __new__(base, value):
            if not value:
                raise MissingError(msg)
            return cls(value)
    return required


class defaultdict(defaultdict):
    __doc__ = defaultdict.__doc__

    def __repr__(self):
        return pprint.pformat(dict(self))


class Missing(object):
    def __nonzero__(self):
        return False


class MissingError(ValueError):
    pass


missing = Missing()


# vim: set ft=python ts=4 sw=4 expandtab :



from repoze.formapi.py24 import defaultdict
from repoze.formapi.py24 import any

class Errors(object):
    """Container for errors.

    Each error will be present in it's `messages` list. Dictionary lookup can
    be used to get at errors which are specific to a field. 

    Note that structure will automatically create entries for non-existing
    keys. This is done to make access from templates etc. easier and less
    fragile.

      >>> from repoze.formapi.error import Errors
      >>> errors = Errors()

    The `errors` object is an object which can easily be converted to unicode
    and that inhibits a dict-like behavior, too.
    
      >>> unicode(errors)
      u''

    Any dict-entry returns a new `errors`-object.
    
      >>> isinstance(errors[None], Errors)
      True

    Errors may be appended to the object using the add operator.
    
      >>> errors += "Abc."
      >>> errors += "Def."

      >>> errors[0]
      'Abc.'

    We can iterate through the errors object.

      >>> tuple(errors)
      ('Abc.', 'Def.')

    The string representation of the object is a concatenation of the
    individual error messages.
    
      >>> unicode(errors)
      u'Abc. Def.'

      >>> len(errors)
      9
      
    The truth value of the errors object is based on the error messages it or
    it's sub errors contain.

      >>> bool(errors)
      True

    If there are no error messages the truth value will be false regardless
    even when there are error keys.

      >>> errors = Errors()
      >>> name_error = errors['name']
      >>> bool(errors)
      False

    Two errors instances are considered equal when they have the same keys with
    the same messages.

      >>> a = Errors()
      >>> a['foo'].append('Error')
      >>> b = Errors()
      >>> b['foo'].append('Error')
      >>> a  == b
      True

    Adding an error to one makes them unequal.

      >>> a['bar'].append('Error')
      >>> a == b
      False

    We can use the standard dictionary ``get`` method.

      >>> a.get('foo')
      <Errors: ['Error'], defaultdict(<class 'repoze.formapi.error.Errors'>, {})>
      >>> a.get('boo', False)
      False

    """

    _messages = _dict = None

    def __init__(self, *args, **kwargs):
        self._dict = defaultdict(Errors, *args, **kwargs)
        self._messages = []

    def __nonzero__(self):
        return bool(self._messages) or any(self._dict.itervalues())

    def __repr__(self):
         return '<Errors: %r, %r>' % (self._messages, self._dict)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._messages[key]
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict
    has_key = __contains__

    def __unicode__(self):
        return u" ".join(self._messages)

    def __str__(self):
        return str(unicode(self))

    def __iter__(self):
        return iter(self._messages)

    def __len__(self):
        return len(unicode(self))

    def __add__(self, error):
        self.append(error)
        return self

    def __getattr__(self, name):
        if name in type(self).__dict__:
            return object.__getattribute__(self, name)
        raise AttributeError(name)

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self._messages == other._messages and self._dict == other._dict

    def append(self, error):
        self._messages.append(error)

    def get(self, key, default=None):
        assert isinstance(key, basestring), "Key must be a string."
        return self._dict.get(key, default)



##############################################################################
#
# Copyright (c) 2008 Agendaless Consulting and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

import os
from setuptools import setup, find_packages, Extension

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(name='repoze.formapi',
      version = '0.6.1',
      description='Minimalistic form library.',
      long_description=(
          read('README.txt') +
          read('src/repoze/formapi/README.txt') +
          read('CHANGES.txt')
          ),
      keywords = "forms html validation marshalling",
      classifiers = [
          'Development Status :: 5 - Production/Stable',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Zope Public License',
          'Programming Language :: Python',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: Internet :: WWW/HTTP'],
      author="Malthe Borch",
      author_email="repoze-dev@lists.repoze.org",
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      packages=find_packages('src'),
      package_dir = {'': 'src'},
      namespace_packages=['repoze', ],
      install_requires=[
          'setuptools',
          ],
      include_package_data = True,
      zip_safe = False,
      test_suite="repoze.formapi.tests.test_suite",
      )
